# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Basic datatypes for parsed pyc files."""

from dataclasses import dataclass

from typing import Any, List, Optional, Tuple, Union


@dataclass
class CodeTypeBase:
    """Pure python types.CodeType with python version added."""

    python_version: Tuple[int, int]
    co_argcount: int
    co_posonlyargcount: int
    co_kwonlyargcount: int
    co_stacksize: int
    co_flags: int
    co_code: bytes
    co_consts: List[object]
    co_names: List[str]
    co_filename: Union[bytes, str]
    co_name: str
    co_firstlineno: int

    def __repr__(self):
        return f"<code: {self.co_name}>"


@dataclass
class CodeType38(CodeTypeBase):
    """CodeType for python 3.8 - 3.10."""

    co_nlocals: int
    co_lnotab: bytes
    co_varnames: List[str]
    co_freevars: Tuple[str, ...]
    co_cellvars: Tuple[str, ...]


@dataclass
class CodeType311(CodeTypeBase):
    """CodeType for python 3.11+."""

    co_qualname: str
    co_localsplusnames: Tuple[str, ...]
    co_localspluskinds: Tuple[int, ...]
    co_linetable: bytes
    co_exceptiontable: bytes


@dataclass
class Opcode:
    """Opcode with names and line numbers."""

    offset: int
    line: int
    endline: Optional[int]
    col: Optional[int]
    endcol: Optional[int]
    op: int
    name: str
    arg: Optional[int]
    argval: Any

    def __str__(self):
        ret = f"{self.line:>5}{self.offset:>6}  {self.name:<30}"
        if self.arg is not None:
            ret += f" {self.arg:>5} {self.argval}"
        return ret


@dataclass
class ExceptionTableEntry:
    """Exception table entry in python 3.11+."""

    start: int
    end: int
    target: int
    depth: int
    lasti: bool

    def pretty_format(self):
        return (
            f"{self.start} to {self.end} -> {self.target} "
            f"[{self.depth}] {'lasti' if self.lasti else ''}"
        )


@dataclass
class ExceptionTable:
    """Exception table in python 3.11+."""

    entries: List[ExceptionTableEntry]

    def __bool__(self):
        return bool(self.entries)


@dataclass
class DisassembledCode:
    """Tree of bytecode and associated opcode list."""

    code: CodeTypeBase
    opcodes: List[Opcode]
    exception_table: ExceptionTable
    children: "List[DisassembledCode]"

    @property
    def name(self):
        return self.code.co_name

    @property
    def python_version(self):
        return self.code.python_version

    def get_child(self, name) -> "Optional[DisassembledCode]":
        """Get the first child with name = `name`."""
        for c in self.children:
            if c.name == name:
                return c

    def pretty_format(self, indent=0):
        header = (
            f"-- <code object {self.name}>, file: {self.code.co_filename}, "
            f"line: {self.code.co_firstlineno}"
        )
        out = [(indent, header)]
        for op in self.opcodes:
            out.append((indent, str(op)))
        if self.exception_table:
            out.append((indent, "ExceptionTable:"))
            for e in self.exception_table.entries:
                out.append((indent + 2, e.pretty_format()))
        for c in self.children:
            out.append((0, ""))
            out.extend(c.pretty_format(indent + 2))
        return out

    def pretty_print(self, indent=0):
        lines = self.pretty_format(indent)
        for i, text in lines:
            print(" " * i, text)



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Line table reader."""

import abc
import dataclasses

from typing import List, Optional

from . import types


@dataclasses.dataclass
class Entry:
    """Position information for an opcode."""

    offset: int
    end_offset: int
    line: int
    # The following are new in 3.11
    endline: Optional[int] = None
    startcol: Optional[int] = None
    endcol: Optional[int] = None


class LineTableReader(abc.ABC):
    """State machine for decoding a pyc line table."""

    def __init__(self, code: types.CodeTypeBase):
        self.lineno = code.co_firstlineno
        self.pos = 0

    @abc.abstractmethod
    def get(self, i: int) -> Entry:
        """Get position information for the instruction at the given position.

        This does NOT allow random access. Call with incremental numbers.

        Args:
          i: The byte position in the bytecode. i needs to stay constant or
          increase between calls.

        Returns:
          The line number corresponding to the position at i.
        """

    @abc.abstractmethod
    def read_all(self) -> List[Entry]:
        """Get all entries in the table.

        Mostly used for testing and debugging purposes.

        Returns:
          A list of entries parsed from the table.
        """


class LnotabReader(LineTableReader):
    """Read code.co_lnotab from pre-Python 3.11."""

    def __init__(self, code: types.CodeType38):
        super().__init__(code)
        self.linetable = code.co_lnotab
        if self.linetable:
            self.end_pos = len(self.linetable)
            self.next_addr = self.linetable[0]
        else:
            self.end_pos = 0
            self.next_addr = 0

    def read_all(self) -> List[Entry]:
        ret = []
        i = 0
        while self.pos < self.end_pos:
            ret.append(self.get(i))
            i = self.next_addr
        return ret


class LineTableReader38(LnotabReader):
    """Read code.co_lnotab from pre-Python 3.10."""

    def get(self, i: int) -> Entry:
        while i >= self.next_addr and self.pos < self.end_pos:
            line_diff = self.linetable[self.pos + 1]
            if line_diff >= 0x80:
                line_diff -= 0x100
            self.lineno += line_diff

            self.pos += 2
            if self.pos < self.end_pos:
                self.next_addr += self.linetable[self.pos]
        return Entry(offset=i, end_offset=self.pos, line=self.lineno)


class LineTableReader310(LnotabReader):
    """Read code.co_lnotab from Python 3.10."""

    def __init__(self, code: types.CodeType38):
        super().__init__(code)
        if self.linetable:
            self.lineno += self.line_delta

    @property
    def line_delta(self):
        line_delta = self.linetable[self.pos + 1]
        if line_delta == 128:
            line_delta = 0
        elif line_delta > 128:
            line_delta -= 256
        return line_delta

    def get(self, i: int) -> Entry:
        while i >= self.next_addr and self.pos < self.end_pos:
            self.pos += 2
            if self.pos < self.end_pos:
                self.next_addr += self.linetable[self.pos]
                self.lineno += self.line_delta
        return Entry(offset=i, end_offset=self.pos, line=self.lineno)


class PyCodeLocation:
    """Enum used in the 3.11 line table reader."""

    INFO_SHORT0 = 0
    INFO_ONE_LINE0 = 10
    INFO_ONE_LINE1 = 11
    INFO_ONE_LINE2 = 12
    INFO_NO_COLUMNS = 13
    INFO_LONG = 14
    INFO_NONE = 15


class LineTableReader311(LineTableReader):
    """Read code.co_linetable for Python 3.11+"""

    def __init__(self, code: types.CodeType311):
        super().__init__(code)
        self.table = code.co_linetable
        self.line = code.co_firstlineno
        self.start = -1
        self.end = 0
        self.end_pos = len(self.table)

    def _read_byte(self):
        b = self.table[self.pos]
        self.pos += 1
        return b

    def _read_varint(self):
        read = self._read_byte()
        val = read & 63
        shift = 0
        while read & 64:
            read = self._read_byte()
            shift += 6
            val |= (read & 63) << shift
        return val

    def _read_signed_varint(self):
        uval = self._read_varint()
        if uval & 1:
            return -(uval >> 1)
        else:
            return uval >> 1

    def read(self):
        b = self._read_byte()
        code = (b >> 3) & 15
        code_units = (b & 7) + 1
        self.start = self.end
        self.end = self.start + code_units * 2
        if code == PyCodeLocation.INFO_NONE:
            endline = -1
            col = -1
            endcol = -1
        elif code == PyCodeLocation.INFO_LONG:
            self.line += self._read_signed_varint()
            endline = self.line + self._read_varint()
            col = self._read_varint() - 1
            endcol = self._read_varint() - 1
        elif code == PyCodeLocation.INFO_NO_COLUMNS:
            self.line += self._read_signed_varint()
            endline = self.line
            col = -1
            endcol = -1
        elif code in (
            PyCodeLocation.INFO_ONE_LINE0,
            PyCodeLocation.INFO_ONE_LINE1,
            PyCodeLocation.INFO_ONE_LINE2,
        ):
            self.line += code - 10
            endline = self.line
            col = self._read_byte()
            endcol = self._read_byte()
        else:
            b2 = self._read_byte()
            assert b2 & 128 == 0
            endline = self.line
            col = (code << 3) | (b2 >> 4)
            endcol = col + (b2 & 15)
        return (endline, col, endcol)

    def get(self, i: int) -> Entry:
        if self.pos >= self.end_pos:
            # Happens for the final RETURN_VALUE in generator expressions
            endline = startcol = endcol = -1
        else:
            endline = startcol = endcol = -1
            # Method calls seem to generate extra linetable entries so we cannot
            # just read one linetable entry per opcode.
            while self.end <= i and self.pos < self.end_pos:
                endline, startcol, endcol = self.read()
        return Entry(
            offset=i,
            end_offset=self.end,
            line=self.line,
            endline=endline,
            startcol=startcol,
            endcol=endcol,
        )

    def read_all(self):
        ret = []
        i = 0
        while self.pos < self.end_pos:
            ret.append(self.get(i))
            i = self.end
        return ret


class ExceptionTableReader:
    """Read the exception table in 3.11+."""

    def __init__(self, code: types.CodeType311):
        self.table = code.co_exceptiontable
        self.pos = 0
        self.end_pos = len(self.table)

    def _read_byte(self):
        b = self.table[self.pos]
        self.pos += 1
        return b

    def _read_varint(self):
        read = self._read_byte()
        val = read & 63
        while read & 64:
            val <<= 6
            read = self._read_byte()
            val |= read & 63
        return val

    def read(self):
        start = self._read_varint() * 2
        length = self._read_varint() * 2
        end = start + length - 2
        target = self._read_varint() * 2
        dl = self._read_varint()
        depth = dl >> 1
        lasti = bool(dl & 1)
        return types.ExceptionTableEntry(
            start=start, end=end, target=target, depth=depth, lasti=lasti
        )

    def read_all(self):
        ret = []
        while self.pos < self.end_pos:
            ret.append(self.read())
        return ret


def linetable_reader(code: types.CodeTypeBase) -> LineTableReader:
    if code.python_version < (3, 10):
        assert isinstance(code, types.CodeType38)
        return LineTableReader38(code)
    elif code.python_version == (3, 10):
        assert isinstance(code, types.CodeType38)
        return LineTableReader310(code)
    else:
        assert isinstance(code, types.CodeType311)
        return LineTableReader311(code)



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A pure python version of marshal.loads."""

import struct
import sys
from typing import Tuple

from . import types


class Type:
    """Type enum from marshal.c."""

    NULL = ord("0")
    NONE = ord("N")
    FALSE = ord("F")
    TRUE = ord("T")
    STOPITER = ord("S")
    ELLIPSIS = ord(".")
    INT = ord("i")
    INT64 = ord("I")
    FLOAT = ord("f")
    BINARY_FLOAT = ord("g")
    COMPLEX = ord("x")
    BINARY_COMPLEX = ord("y")
    LONG = ord("l")
    STRING = ord("s")
    STRINGREF = ord("R")
    INTERNED = ord("t")
    REF = ord("r")
    TUPLE = ord("(")
    LIST = ord("[")
    DICT = ord("{")
    CODE = ord("c")
    UNICODE = ord("u")
    UNKNOWN = ord("?")
    SET = ord("<")
    FROZENSET = ord(">")
    ASCII = ord("a")
    ASCII_INTERNED = ord("A")
    SMALL_TUPLE = ord(")")
    SHORT_ASCII = ord("z")
    SHORT_ASCII_INTERNED = ord("Z")


class Flags:
    """Masks and values used by FORMAT_VALUE opcode."""

    FVC_MASK = 0x3
    FVC_NONE = 0x0
    FVC_STR = 0x1
    FVC_REPR = 0x2
    FVC_ASCII = 0x3
    FVS_MASK = 0x4
    FVS_HAVE_SPEC = 0x4

    # Flag used by CALL_FUNCTION_EX
    CALL_FUNCTION_EX_HAS_KWARGS = 0x1

    # Flags used by MAKE_FUNCTION
    MAKE_FUNCTION_HAS_POS_DEFAULTS = 0x1
    MAKE_FUNCTION_HAS_KW_DEFAULTS = 0x2
    MAKE_FUNCTION_HAS_ANNOTATIONS = 0x4
    MAKE_FUNCTION_HAS_FREE_VARS = 0x8

    # Or-ing this flag to one of the codes above will cause the decoded value to
    # be stored in a reference table for later lookup.
    REF = 0x80

    # Cell kinds
    CO_FAST_LOCAL = 0x20
    CO_FAST_CELL = 0x40
    CO_FAST_FREE = 0x80

    # for co_flags:
    CO_OPTIMIZED = 0x0001
    CO_NEWLOCALS = 0x0002
    CO_VARARGS = 0x0004
    CO_VARKEYWORDS = 0x0008
    CO_NESTED = 0x0010
    CO_GENERATOR = 0x0020
    CO_NOFREE = 0x0040
    CO_COROUTINE = 0x0080
    CO_ITERABLE_COROUTINE = 0x0100
    CO_ASYNC_GENERATOR = 0x0200
    CO_FUTURE_DIVISION = 0x2000
    CO_FUTURE_ABSOLUTE_IMPORT = 0x4000
    CO_FUTURE_WITH_STATEMENT = 0x8000
    CO_FUTURE_PRINT_FUNCTION = 0x10000
    CO_FUTURE_UNICODE_LITERALS = 0x20000


NULL = object()  # sentinel marker


class MarshalReader:
    """Stateful loader for marshalled files."""

    def __init__(self, data: bytes, python_version: Tuple[int, int]):
        self.bufstr = data
        self.bufpos = 0
        self.python_version = python_version
        self.refs = []
        self._stringtable = []

    def eof(self):
        """Return True if we reached the end of the stream."""
        return self.bufpos == len(self.bufstr)

    def load(self):
        """Load an encoded Python data structure."""
        c = ord("?")  # make pylint happy
        try:
            c = self._read_byte()
            if c & Flags.REF:
                # This element might recursively contain other elements, which
                # themselves store things in the refs table. So we need to
                # determine the index position *before* reading the contents of
                # this element.
                idx = self._reserve_ref()
                result = MarshalReader._DISPATCH[c & ~Flags.REF](self)
                self.refs[idx] = result
            else:
                result = MarshalReader._DISPATCH[c](self)
            return result
        except KeyError as e:
            raise ValueError(f"bad marshal code: {chr(c)!r} ({c:02x})") from e
        except IndexError as e:
            raise EOFError() from e

    def _read(self, n):
        """Read n bytes as a string."""
        pos = self.bufpos
        self.bufpos += n
        if self.bufpos > len(self.bufstr):
            raise EOFError()
        return self.bufstr[pos : self.bufpos]

    def _read_byte(self):
        """Read an unsigned byte."""
        pos = self.bufpos
        self.bufpos += 1
        return self.bufstr[pos]

    def _read_short(self):
        """Read a signed 16 bit word."""
        lo = self._read_byte()
        hi = self._read_byte()
        x = lo | (hi << 8)
        if x & 0x8000:
            # sign extension
            x -= 0x10000
        return x

    def _read_long(self):
        """Read a signed 32 bit word."""
        b = self._read(4)
        x = b[0] | b[1] << 8 | b[2] << 16 | b[3] << 24
        if b[3] & 0x80 and x > 0:
            # sign extension
            x = -((1 << 32) - x)
            return int(x)
        else:
            return x

    def _read_long64(self):
        """Read a signed 64 bit integer."""
        b = self._read(8)
        x = (
            b[0]
            | b[1] << 8
            | b[2] << 16
            | b[3] << 24
            | b[4] << 32
            | b[5] << 40
            | b[6] << 48
            | b[7] << 56
        )
        if b[7] & 0x80 and x > 0:
            # sign extension
            x = -((1 << 64) - x)
        return x

    def _read_sized(self):
        """Read a size and a variable number of bytes."""
        n = self._read_long()
        return self._read(n)

    def _reserve_ref(self):
        """Reserve one entry in the reference table.

        This is done before reading an element, because reading an element and
        all its subelements might change the size of the reference table.

        Returns:
          Reserved index position in the reference table.
        """
        # See r_ref_reserve in Python-3.4/Python/marshal.c
        idx = len(self.refs)
        self.refs.append(None)
        return idx

    # pylint: disable=missing-docstring
    # This is a bunch of small methods with self-explanatory names.

    def load_null(self):
        return NULL

    def load_none(self):
        return None

    def load_true(self):
        return True

    def load_false(self):
        return False

    def load_stopiter(self):
        return StopIteration

    def load_ellipsis(self):
        return Ellipsis

    def load_int(self):
        return self._read_long()

    def load_int64(self):
        return self._read_long64()

    def load_long(self):
        """Load a variable length integer."""
        size = self._read_long()
        x = 0
        for i in range(abs(size)):
            d = self._read_short()
            x |= d << (i * 15)
        return x if size >= 0 else -x

    def load_float(self):
        n = self._read_byte()
        s = self._read(n)
        return float(s)

    def load_binary_float(self):
        binary = self._read(8)
        return struct.unpack("<d", binary)[0]

    def load_complex(self):
        real = self.load_float()
        imag = self.load_float()
        return complex(real, imag)

    def load_binary_complex(self):
        binary = self._read(16)
        return complex(*struct.unpack("dd", binary))

    def load_string(self):
        s = self._read_sized()
        return bytes(s)

    def load_interned(self):
        s = self._read_sized()
        ret = sys.intern(s.decode("utf-8", "strict"))
        self._stringtable.append(ret)
        return ret

    def load_stringref(self):
        n = self._read_long()
        return self._stringtable[n]

    def load_unicode(self):
        # We need to convert bytes to a unicode string.
        # We use the 'backslashreplace' error mode in order to handle non-utf8
        # backslash-escaped string literals correctly.
        s = self._read_sized()
        return s.decode("utf8", "backslashreplace")

    def load_ascii(self):
        s = self._read_sized()
        return s.decode("ascii")

    def load_short_ascii(self):
        n = self._read_byte()
        s = self._read(n)
        return s.decode("ascii")

    def load_list(self):
        n = self._read_long()
        return [self.load() for _ in range(n)]

    def load_tuple(self):
        return tuple(self.load_list())

    def load_small_tuple(self):
        n = self._read_byte()
        elts = [self.load() for _ in range(n)]
        return tuple(elts)

    def load_dict(self):
        d = {}
        while True:
            key = self.load()
            if key is NULL:
                break
            value = self.load()
            d[key] = value
        return d

    def load_set(self):
        return set(self.load_list())

    def load_frozenset(self):
        return frozenset(self.load_list())

    def load_ref(self):
        n = self._read_long()
        return self.refs[n]

    def load_code(self):
        if self.python_version < (3, 11):
            return self.load_code_3_8()
        else:
            return self.load_code_3_11()

    def load_code_3_8(self):
        """Load a Python code object."""
        argcount = self._read_long()
        posonlyargcount = self._read_long()
        kwonlyargcount = self._read_long()
        nlocals = self._read_long()
        stacksize = self._read_long()
        flags = self._read_long()
        code = self.load()
        consts = self.load()
        names = self.load()
        varnames = self.load()
        freevars = self.load()
        cellvars = self.load()
        filename = self.load()
        name = self.load()
        firstlineno = self._read_long()
        # lnotab, from
        # https://github.com/python/cpython/blob/master/Objects/lnotab_notes.txt:
        # 'an array of unsigned bytes disguised as a Python bytes object'.
        lnotab = self.load()
        return types.CodeType38(
            co_argcount=argcount,
            co_posonlyargcount=posonlyargcount,
            co_kwonlyargcount=kwonlyargcount,
            co_nlocals=nlocals,
            co_stacksize=stacksize,
            co_flags=flags,
            co_code=code,
            co_consts=consts,
            co_names=names,
            co_varnames=varnames,
            co_filename=filename,
            co_name=name,
            co_firstlineno=firstlineno,
            co_lnotab=lnotab,
            co_freevars=freevars,
            co_cellvars=cellvars,
            python_version=self.python_version,
        )

    def load_code_3_11(self):
        """Load a Python code object."""
        argcount = self._read_long()
        posonlyargcount = self._read_long()
        kwonlyargcount = self._read_long()
        stacksize = self._read_long()
        flags = self._read_long()
        code = self.load()
        consts = self.load()
        names = self.load()
        localsplusnames = self.load()
        localspluskinds = self.load()
        filename = self.load()
        name = self.load()
        qualname = self.load()
        firstlineno = self._read_long()
        linetable = self.load()
        exceptiontable = self.load()

        return types.CodeType311(
            co_argcount=argcount,
            co_posonlyargcount=posonlyargcount,
            co_kwonlyargcount=kwonlyargcount,
            co_stacksize=stacksize,
            co_flags=flags,
            co_code=code,
            co_consts=consts,
            co_names=names,
            co_localsplusnames=localsplusnames,
            co_localspluskinds=localspluskinds,
            co_filename=filename,
            co_name=name,
            co_qualname=qualname,
            co_firstlineno=firstlineno,
            co_linetable=linetable,
            co_exceptiontable=exceptiontable,
            python_version=self.python_version,
        )

    # pylint: enable=missing-docstring

    _DISPATCH = {
        Type.ASCII: load_ascii,
        Type.ASCII_INTERNED: load_ascii,
        Type.BINARY_COMPLEX: load_binary_complex,
        Type.BINARY_FLOAT: load_binary_float,
        Type.CODE: load_code,
        Type.COMPLEX: load_complex,
        Type.DICT: load_dict,
        Type.ELLIPSIS: load_ellipsis,
        Type.FALSE: load_false,
        Type.FLOAT: load_float,
        Type.FROZENSET: load_frozenset,
        Type.INT64: load_int64,
        Type.INT: load_int,
        Type.INTERNED: load_interned,
        Type.LIST: load_list,
        Type.LONG: load_long,
        Type.NONE: load_none,
        Type.NULL: load_null,
        Type.REF: load_ref,
        Type.SET: load_set,
        Type.SHORT_ASCII: load_short_ascii,
        Type.SHORT_ASCII_INTERNED: load_short_ascii,
        Type.SMALL_TUPLE: load_small_tuple,
        Type.STOPITER: load_stopiter,
        Type.STRING: load_string,
        Type.STRINGREF: load_stringref,
        Type.TRUE: load_true,
        Type.TUPLE: load_tuple,
        Type.UNICODE: load_unicode,
    }


def loads(data: bytes, python_version: Tuple[int, int]):
    um = MarshalReader(data, python_version)
    result = um.load()
    if not um.eof():
        leftover = um.bufstr[um.bufpos :]
        if len(leftover) > 80:
            raise BufferError(
                f"trailing bytes in marshal data ({um.bufpos}...)"
            )
        else:
            raise BufferError(
                f"trailing bytes in marshal data ({um.bufpos}...): {leftover}"
            )
    return result



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for marshal.py."""

import unittest

from pycnite import marshal


class Base(unittest.TestCase):
    """Base class for marshal reader tests."""

    def assertStrictEqual(self, s1, s2):
        self.assertEqual(s1, s2)
        self.assertEqual(type(s1), type(s2))

    def load(self, s, python_version=None):
        python_version = python_version or (3, 9)
        return marshal.loads(s, python_version)


class TestMarshalReader(Base):
    """Tests for marshal reader."""

    def test_load_none(self):
        self.assertIsNone(self.load(b"N"))

    def test_load_false(self):
        self.assertEqual(self.load(b"F"), False)

    def test_load_true(self):
        self.assertEqual(self.load(b"T"), True)

    def test_load_stopiter(self):
        self.assertEqual(self.load(b"S"), StopIteration)

    def test_load_ellipsis(self):
        self.assertEqual(self.load(b"."), Ellipsis)

    def test_load_int(self):
        self.assertEqual(self.load(b"i\1\2\3\4"), 0x04030201)
        self.assertEqual(self.load(b"i\xff\xff\xff\xff"), -1)

    def test_load_int64(self):
        self.assertEqual(self.load(b"I\1\2\3\4\5\6\7\x08"), 0x0807060504030201)
        self.assertEqual(self.load(b"I\xff\xff\xff\xff\xff\xff\xff\xff"), -1)

    def test_load_float(self):
        self.assertEqual(self.load(b"f\x040.25"), 0.25)

    def test_load_long_float(self):
        self.assertEqual(self.load(b"f\xff0." + (b"9" * 253)), 1.0)

    def test_load_binary_float(self):
        self.assertEqual(self.load(b"g\0\0\0\0\0\0\xd0\x3f"), 0.25)

    def test_load_complex(self):
        self.assertEqual(self.load(b"x\3.25\3.25"), 0.25 + 0.25j)

    def test_load_binary_complex(self):
        c = self.load(b"y\0\0\0\0\0\0\xf0\x3f\0\0\0\0\0\0\xf0\x3f")
        self.assertEqual(c, 1 + 1j)

    def test_load_long(self):
        """Load a variable length integer."""
        self.assertEqual(
            self.load(b"l\3\0\0\0\1\0\2\0\3\0"), 1 + 2 * 2**15 + 3 * 2**30
        )
        self.assertEqual(self.load(b"l\xff\xff\xff\xff\1\0"), -1)
        self.assertEqual(self.load(b"l\xfe\xff\xff\xff\1\0\2\0"), -65537)

    def test_load_string(self):
        self.assertStrictEqual(
            self.load(b"s\4\0\0\0test", (3, 9)), bytes(b"test")
        )

    def test_load_interned(self):
        self.assertStrictEqual(self.load(b"t\4\0\0\0test"), "test")

    def test_load_stringref(self):
        st = (
            b"(\4\0\0\0"  # tuple of 4
            b"t\4\0\0\0abcd"  # store "abcd" at 0
            b"t\4\0\0\0efgh"  # store "efgh" at 1
            b"R\0\0\0\0"  # retrieve stringref 0
            b"R\1\0\0\0"
        )  # retrieve stringref 1
        self.assertEqual(self.load(st), ("abcd", "efgh", "abcd", "efgh"))

    def test_load_tuple(self):
        self.assertEqual(self.load(b"(\2\0\0\0TF"), (True, False))

    def test_load_list(self):
        self.assertEqual(self.load(b"[\2\0\0\0TF"), [True, False])

    def test_load_dict(self):
        self.assertEqual(self.load(b"{TFFN0"), {True: False, False: None})

    def test_load_unicode(self):
        self.assertStrictEqual(self.load(b"u\4\0\0\0test", (3, 9)), "test")
        # This character is \u00e4 (umlaut a).
        s = b"u\2\0\0\0\xc3\xa4"
        self.assertStrictEqual(self.load(s, (3, 9)), "\xe4")

    def test_load_set(self):
        self.assertEqual(self.load(b"<\3\0\0\0FTN"), {True, False, None})

    def test_load_frozenset(self):
        self.assertEqual(
            self.load(b">\3\0\0\0FTN"), frozenset([True, False, None])
        )

    def test_load_ref(self):
        data = (
            b"(\4\0\0\0"  # tuple of 4
            b"\xe9\0\1\2\3"  # store 0x03020100 at 0
            b"\xe9\4\5\6\7"  # store 0x07060504 at 1
            b"r\0\0\0\0"  # retrieve 0
            b"r\1\0\0\0"
        )  # retrieve 1
        self.assertEqual(
            self.load(data), (0x03020100, 0x7060504, 0x03020100, 0x7060504)
        )

    def test_load_ascii(self):
        self.assertStrictEqual(self.load(b"a\4\0\0\0test"), "test")

    def test_load_ascii_interned(self):
        self.assertStrictEqual(self.load(b"A\4\0\0\0test"), "test")

    def test_load_small_tuple(self):
        self.assertEqual(self.load(b")\2TF"), (True, False))

    def test_load_short_ascii(self):
        self.assertStrictEqual(self.load(b"z\4test"), "test")

    def test_load_short_ascii_interned(self):
        self.assertStrictEqual(self.load(b"Z\4test"), "test")

    def test_truncated(self):
        self.assertRaises(EOFError, lambda: self.load(b"f\x020"))

    def test_trailing(self):
        self.assertRaises(BufferError, lambda: self.load(b"N\0"))

    def test_illegal(self):
        self.assertRaises(ValueError, lambda: self.load(b"\7"))

    def test_truncated_byte(self):
        self.assertRaises(EOFError, lambda: self.load(b"f"))


class TestCodeReader(Base):
    """Tests for marshal code reader."""

    def test_code_3_8(self):
        # Code from marshal.dumps of
        #   def f():
        #     x = 10
        #     def g():
        #       return x
        #     return g
        bytecode = (
            b"\xe3\x01\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x03\x00\x00"
            b"\x00\x03\x00\x00\x00\x03\x00\x00\x00s\x14\x00\x00\x00d\x01\x89"
            b"\x00\x87\x00f\x01d\x02d\x03\x84\x08}\x02|\x02S\x00)\x04N\xe9\n"
            b"\x00\x00\x00c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x01\x00\x00\x00\x13\x00\x00\x00s\x04\x00\x00"
            b"\x00\x88\x00S\x00)\x01N\xa9\x00r\x02\x00\x00\x00\xa9\x01\xda"
            b"\x01xr\x02\x00\x00\x00\xfa\x1e<ipython-input-1-fe7579e34ce2>"
            b"\xda\x01g\x03\x00\x00\x00s\x02\x00\x00\x00\x04\x01z\x0c"
            b"f.<locals>.gr\x02\x00\x00\x00)\x03\xda\x01a\xda\x01br\x06\x00"
            b"\x00\x00r\x02\x00\x00\x00r\x03\x00\x00\x00r\x05\x00\x00\x00\xda"
            b"\x01f\x01\x00\x00\x00s\x06\x00\x00\x00\x04\x01\x0c\x01\x04\x02"
        )
        code = marshal.loads(bytecode, (3, 10))
        self.assertEqual(code.co_argcount, 1)
        self.assertEqual(code.co_kwonlyargcount, 1)
        self.assertEqual(code.co_nlocals, 3)
        self.assertEqual(code.co_stacksize, 3)
        self.assertEqual(code.co_flags, 3)
        self.assertEqual(code.co_names, ())
        self.assertEqual(
            code.co_varnames,
            (
                "a",
                "b",
                "g",
            ),
        )
        self.assertEqual(code.co_freevars, ())
        self.assertEqual(code.co_cellvars, ("x",))

    def test_code_3_11(self):
        # Code from marshal.dumps of
        #   def f(a, *, b):
        #     x = 10
        #     def g():
        #       return x
        #     return g
        bytecode = (
            b"\xe3\x01\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x02\x00"
            b"\x00\x00\x03\x00\x00\x00\xf3\x16\x00\x00\x00\x87\x03\x97\x00d"
            b"\x01\x8a\x03\x88\x03f\x01d\x02\x84\x08}\x02|\x02S\x00)\x03N\xe9"
            b"\n\x00\x00\x00c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x01\x00\x00\x00\x13\x00\x00\x00\xf3\x08\x00\x00\x00\x95\x01\x97"
            b"\x00\x89\x00S\x00)\x01N\xa9\x00)\x01\xda\x01xs\x01\x00\x00\x00"
            b"\x80\xfa\x1e<ipython-input-2-fe7579e34ce2>\xda\x01gz\x0c"
            b"f.<locals>.g\x03\x00\x00\x00s\x08\x00\x00\x00\xf8\x80\x00\xd8"
            b"\x0f\x10\x88\x08\xf3\x00\x00\x00\x00r\x04\x00\x00\x00)\x04\xda"
            b"\x01a\xda\x01br\x07\x00\x00\x00r\x05\x00\x00\x00s\x04\x00\x00"
            b"\x00   @r\x06\x00\x00\x00\xda\x01fr\x0b\x00\x00\x00\x01\x00\x00"
            b"\x00s&\x00\x00\x00\xf8\x80\x00\xd8\x08\n\x80A\xf0\x02\x01\x05"
            b"\x11\xf0\x00\x01\x05\x11\xf0\x00\x01\x05\x11\xf0\x00\x01\x05\x11"
            b"\xf0\x00\x01\x05\x11\xe0\x0b\x0c\x80Hr\x08\x00\x00\x00"
        )
        code = marshal.loads(bytecode, (3, 11))
        self.assertEqual(code.co_argcount, 1)
        self.assertEqual(code.co_kwonlyargcount, 1)
        self.assertEqual(code.co_stacksize, 2)
        self.assertEqual(code.co_flags, 3)
        self.assertEqual(code.co_names, ())


if __name__ == "__main__":
    unittest.main()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for pycnite.magic."""

import unittest

from . import base
from pycnite import magic


class TestMagic(unittest.TestCase):
    """Test magic number parsing."""

    def test_version(self):
        for version in base.VERSIONS:
            path = base.test_pyc("basic", version)
            with open(path, "rb") as f:
                magic_number = f.read(2)
            self.assertEqual(
                version, magic.magic_number_to_version(magic_number)
            )


if __name__ == "__main__":
    unittest.main()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Bytecode reader."""

from dataclasses import dataclass

from typing import cast, Iterable, List, Optional

from . import linetable
from . import mapping
from . import types


# From cpython/Include/opcodes.h
HAVE_ARGUMENT = 90
EXTENDED_ARG = 144


@dataclass
class RawOpcode:
    """Opcode parsed from code.co_code."""

    start: int
    end: int
    op: int
    arg: Optional[int]


def wordcode_reader(data: bytes) -> Iterable[RawOpcode]:
    """Reads binary data from pyc files."""

    extended_arg = 0
    start = 0
    for pos in range(0, len(data), 2):
        op = data[pos]
        if op == EXTENDED_ARG:
            oparg = data[pos + 1] | extended_arg
            extended_arg = oparg << 8
        elif op >= HAVE_ARGUMENT:
            oparg = data[pos + 1] | extended_arg
            extended_arg = 0
        else:
            oparg = None
            extended_arg = 0
        # Don't yield EXTENDED_ARG; it is part of the next opcode.
        if op != EXTENDED_ARG:
            yield RawOpcode(start, pos + 2, op, oparg)
            start = pos + 2


# In python 3.11+ attributes like co_consts can be None, so following
# cpython/Lib/dis.py we define an UNKNOWN value for args whose value we cannot
# calculate.
class _Unknown:
    def __repr__(self):
        return "<unknown>"


# Sentinel to represent values that cannot be calculated
UNKNOWN = _Unknown()


def _is_backward_jump(name):
    return "JUMP_BACKWARD" in name


class Disassembler:
    """Disassemble code."""

    def __init__(self, code):
        self.code = code
        self.python_version = code.python_version[:2]
        self.opmap = mapping.get_mapping(self.python_version)
        if self.python_version < (3, 11):
            code = cast(types.CodeType38, code)
            self.cell_names = code.co_cellvars + code.co_freevars

    def _lookup(self, vals: Optional[tuple], arg: int):
        if vals is None or arg >= len(vals):
            return UNKNOWN
        return vals[arg]

    def _decode_arg(self, name: str, arg_type: int, arg: int, end_pos: int):
        if self.python_version >= (3, 12):
            if name == "COMPARE_OP":
                return arg >> 4
            if name == "LOAD_ATTR":
                return arg >> 1
            if name == "LOAD_SUPER_ATTR":
                return arg >> 2
        if self.python_version >= (3, 11):
            if name == "LOAD_GLOBAL":
                return arg // 2
        if self.python_version >= (3, 10):
            if arg_type == mapping.JREL:
                signed_arg = -arg if _is_backward_jump(name) else arg
                target = signed_arg * 2 + end_pos
                if self.python_version >= (3, 12):
                    target += 2 * mapping.caches(name, self.python_version)
                return target
            elif arg_type == mapping.JABS:
                return arg * 2
        if arg_type == mapping.JREL:
            return arg + end_pos
        return arg

    def _get_argval(self, name: str, arg: int, end_pos: int):
        arg_type = mapping.arg_type(name, self.python_version)
        arg = self._decode_arg(name, arg_type, arg, end_pos)
        if arg_type == mapping.CONST:
            return self._lookup(self.code.co_consts, arg)
        elif arg_type == mapping.NAME:
            return self._lookup(self.code.co_names, arg)
        elif arg_type == mapping.LOCAL:
            if self.python_version >= (3, 11):
                return self._lookup(self.code.co_localsplusnames, arg)
            else:
                return self._lookup(self.code.co_varnames, arg)
        elif arg_type == mapping.FREE:
            if self.python_version >= (3, 11):
                return self._lookup(self.code.co_localsplusnames, arg)
            else:
                return self._lookup(self.cell_names, arg)
        elif name == "CALL_INTRINSIC_1":
            return mapping.intrinsic_1_desc(arg, self.python_version)
        elif name == "CALL_INTRINSIC_2":
            return mapping.intrinsic_2_desc(arg, self.python_version)
        else:
            return arg

    def dis(self) -> List[types.Opcode]:
        """Disassemble code."""
        lt = linetable.linetable_reader(self.code)
        ret = []
        for o in wordcode_reader(self.code.co_code):
            if o.op == 0:  # CACHE
                continue
            name = self.opmap[o.op]
            pos = lt.get(o.start)
            argval = self._get_argval(name, o.arg, o.end)
            if isinstance(argval, types.CodeTypeBase):
                argval = f"<code:{argval.co_name}>"
            op = types.Opcode(
                offset=o.start,
                name=name,
                line=pos.line,
                endline=pos.endline,
                col=pos.startcol,
                endcol=pos.endcol,
                op=o.op,
                arg=o.arg,
                argval=argval,
            )
            ret.append(op)
        return ret


def dis(code: types.CodeTypeBase) -> List[types.Opcode]:
    """Disassemble a single piece of top-level code."""
    return Disassembler(code).dis()


def dis_all(code: types.CodeTypeBase) -> types.DisassembledCode:
    """Recursively disassemble code and contained code blocks."""
    opcodes = dis(code)
    if code.python_version >= (3, 11):
        code = cast(types.CodeType311, code)
        et = linetable.ExceptionTableReader(code)
        exc_table = types.ExceptionTable(et.read_all())
    else:
        # Setting this to None will complicate typechecking without adding any
        # real safety; client code will typically check for version >= 3.11
        # before accessing the exception table, rather than doing type or
        # presence checks on the attribute.
        exc_table = types.ExceptionTable([])
    ret = types.DisassembledCode(
        code=code, opcodes=opcodes, exception_table=exc_table, children=[]
    )
    for child in code.co_consts:
        if hasattr(child, "co_code"):
            ret.children.append(dis_all(child))
    return ret



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mapping of opcode codes to names."""

from typing import Dict, Mapping, Optional, Tuple

OpMap = Dict[int, str]
Overlay = Dict[int, Optional[str]]
OpArgs = Mapping[str, int]
OpCaches = Mapping[str, int]


def _overlay_mapping(mapping: OpMap, new_entries: Overlay) -> OpMap:
    ret: Overlay = mapping.copy()
    ret.update(new_entries)
    return {k: v for k, v in ret.items() if v is not None}


PYTHON_3_8_MAPPING: OpMap = {
    1: "POP_TOP",
    2: "ROT_TWO",
    3: "ROT_THREE",
    4: "DUP_TOP",
    5: "DUP_TOP_TWO",
    6: "ROT_FOUR",  # ROT_FOUR returns under a different, cooler id!
    9: "NOP",
    10: "UNARY_POSITIVE",
    11: "UNARY_NEGATIVE",
    12: "UNARY_NOT",
    15: "UNARY_INVERT",
    16: "BINARY_MATRIX_MULTIPLY",
    17: "INPLACE_MATRIX_MULTIPLY",
    19: "BINARY_POWER",
    20: "BINARY_MULTIPLY",
    22: "BINARY_MODULO",
    23: "BINARY_ADD",
    24: "BINARY_SUBTRACT",
    25: "BINARY_SUBSCR",
    26: "BINARY_FLOOR_DIVIDE",
    27: "BINARY_TRUE_DIVIDE",
    28: "INPLACE_FLOOR_DIVIDE",
    29: "INPLACE_TRUE_DIVIDE",
    50: "GET_AITER",
    51: "GET_ANEXT",
    52: "BEFORE_ASYNC_WITH",
    53: "BEGIN_FINALLY",
    54: "END_ASYNC_FOR",
    55: "INPLACE_ADD",
    56: "INPLACE_SUBTRACT",
    57: "INPLACE_MULTIPLY",
    59: "INPLACE_MODULO",
    60: "STORE_SUBSCR",
    61: "DELETE_SUBSCR",
    62: "BINARY_LSHIFT",
    63: "BINARY_RSHIFT",
    64: "BINARY_AND",
    65: "BINARY_XOR",
    66: "BINARY_OR",
    67: "INPLACE_POWER",
    68: "GET_ITER",
    69: "GET_YIELD_FROM_ITER",
    70: "PRINT_EXPR",
    71: "LOAD_BUILD_CLASS",
    72: "YIELD_FROM",
    73: "GET_AWAITABLE",
    75: "INPLACE_LSHIFT",
    76: "INPLACE_RSHIFT",
    77: "INPLACE_AND",
    78: "INPLACE_XOR",
    79: "INPLACE_OR",
    81: "WITH_CLEANUP_START",
    82: "WITH_CLEANUP_FINISH",
    83: "RETURN_VALUE",
    84: "IMPORT_STAR",
    85: "SETUP_ANNOTATIONS",
    86: "YIELD_VALUE",
    87: "POP_BLOCK",
    88: "END_FINALLY",
    89: "POP_EXCEPT",
    90: "STORE_NAME",
    91: "DELETE_NAME",
    92: "UNPACK_SEQUENCE",
    93: "FOR_ITER",
    94: "UNPACK_EX",
    95: "STORE_ATTR",
    96: "DELETE_ATTR",
    97: "STORE_GLOBAL",
    98: "DELETE_GLOBAL",
    100: "LOAD_CONST",
    101: "LOAD_NAME",
    102: "BUILD_TUPLE",
    103: "BUILD_LIST",
    104: "BUILD_SET",
    105: "BUILD_MAP",
    106: "LOAD_ATTR",
    107: "COMPARE_OP",
    108: "IMPORT_NAME",
    109: "IMPORT_FROM",
    110: "JUMP_FORWARD",
    111: "JUMP_IF_FALSE_OR_POP",
    112: "JUMP_IF_TRUE_OR_POP",
    113: "JUMP_ABSOLUTE",
    114: "POP_JUMP_IF_FALSE",
    115: "POP_JUMP_IF_TRUE",
    116: "LOAD_GLOBAL",
    122: "SETUP_FINALLY",
    124: "LOAD_FAST",
    125: "STORE_FAST",
    126: "DELETE_FAST",
    130: "RAISE_VARARGS",
    131: "CALL_FUNCTION",
    132: "MAKE_FUNCTION",
    133: "BUILD_SLICE",
    134: "MAKE_CLOSURE",
    135: "LOAD_CLOSURE",
    136: "LOAD_DEREF",
    137: "STORE_DEREF",
    138: "DELETE_DEREF",
    141: "CALL_FUNCTION_KW",
    142: "CALL_FUNCTION_EX",
    143: "SETUP_WITH",
    144: "EXTENDED_ARG",
    145: "LIST_APPEND",
    146: "SET_ADD",
    147: "MAP_ADD",
    148: "LOAD_CLASSDEREF",
    149: "BUILD_LIST_UNPACK",
    150: "BUILD_MAP_UNPACK",
    151: "BUILD_MAP_UNPACK_WITH_CALL",
    152: "BUILD_TUPLE_UNPACK",
    153: "BUILD_SET_UNPACK",
    154: "SETUP_ASYNC_WITH",
    155: "FORMAT_VALUE",
    156: "BUILD_CONST_KEY_MAP",
    157: "BUILD_STRING",
    158: "BUILD_TUPLE_UNPACK_WITH_CALL",
    160: "LOAD_METHOD",
    161: "CALL_METHOD",
    162: "CALL_FINALLY",
    163: "POP_FINALLY",
}

PYTHON_3_9_MAPPING: OpMap = _overlay_mapping(
    PYTHON_3_8_MAPPING,
    {
        48: "RERAISE",
        49: "WITH_EXCEPT_START",
        53: None,  # was BEGIN_FINALLY in 3.8
        74: "LOAD_ASSERTION_ERROR",
        81: None,  # was WITH_CLEANUP_START in 3.8
        82: "LIST_TO_TUPLE",  # was WITH_CLEANUP_FINISH in 3.8
        88: None,  # was END_FINALLY in 3.8
        117: "IS_OP",
        118: "CONTAINS_OP",
        121: "JUMP_IF_NOT_EXC_MATCH",
        149: None,  # was BUILD_LIST_UNPACK in 3.8
        150: None,  # was BUILD_MAP_UNPACK in 3.8
        151: None,  # was BUILD_MAP_UNPACK_WITH_CALL in 3.8
        152: None,  # was BUILD_TUPLE_UNPACK in 3.8
        153: None,  # was BUILD_SET_UNPACK in 3.8
        158: None,  # was BUILD_TUPLE_UNPACK_WITH_CALL in 3.8
        162: "LIST_EXTEND",  # was CALL_FINALLY in 3.8
        163: "SET_UPDATE",  # was POP_FINALLY in 3.8
        164: "DICT_MERGE",
        165: "DICT_UPDATE",
    },
)

PYTHON_3_10_MAPPING: OpMap = _overlay_mapping(
    PYTHON_3_9_MAPPING,
    {
        30: "GET_LEN",
        31: "MATCH_MAPPING",
        32: "MATCH_SEQUENCE",
        33: "MATCH_KEYS",
        34: "COPY_DICT_WITHOUT_KEYS",
        48: None,  # was RERAISE in 3.9
        99: "ROT_N",
        119: "RERAISE",
        129: "GEN_START",
        152: "MATCH_CLASS",
    },
)

PYTHON_3_11_MAPPING: OpMap = _overlay_mapping(
    PYTHON_3_10_MAPPING,
    {
        0: "CACHE",
        2: "PUSH_NULL",  # was ROT_TWO in 3.10
        3: None,  # was ROT_THREE in 3.10
        4: None,  # was DUP_TOP in 3.10
        5: None,  # was DUP_TOP_TWO in 3.10
        6: None,  # was ROT_FOUR in 3.10
        16: None,  # was BINARY_MATRIX_MULTIPLY in 3.10
        17: None,  # was INPLACE_MATRIX_MULTIPLY in 3.10
        19: None,  # was BINARY_POWER in 3.10
        20: None,  # was BINARY_MULTIPLY in 3.10
        22: None,  # was BINARY_MODULO in 3.10
        23: None,  # was BINARY_ADD in 3.10
        24: None,  # was BINARY_SUBTRACT in 3.10
        26: None,  # was BINARY_FLOOR_DIVIDE in 3.10
        27: None,  # was BINARY_TRUE_DIVIDE in 3.10
        28: None,  # was INPLACE_FLOOR_DIVIDE in 3.10
        29: None,  # was INPLACE_TRUE_DIVIDE in 3.10
        34: None,  # was COPY_DICT_WITHOUT_KEYS in 3.10
        35: "PUSH_EXC_INFO",
        36: "CHECK_EXC_MATCH",
        37: "CHECK_EG_MATCH",
        53: "BEFORE_WITH",
        55: None,  # was INPLACE_ADD in 3.10
        56: None,  # was INPLACE_SUBTRACT in 3.10
        57: None,  # was INPLACE_MULTIPLY in 3.10
        59: None,  # was INPLACE_MODULO in 3.10
        62: None,  # was BINARY_LSHIFT in 3.10
        63: None,  # was BINARY_RSHIFT in 3.10
        64: None,  # was BINARY_AND in 3.10
        65: None,  # was BINARY_XOR in 3.10
        66: None,  # was BINARY_OR in 3.10
        67: None,  # was INPLACE_POWER in 3.10
        72: None,  # was YIELD_FROM in 3.10
        73: None,  # was GET_AWAITABLE in 3.10
        75: "RETURN_GENERATOR",  # was INPLACE_LSHIFT in 3.10
        76: None,  # was INPLACE_RSHIFT in 3.10
        77: None,  # was INPLACE_AND in 3.10
        78: None,  # was INPLACE_XOR in 3.10
        79: None,  # was INPLACE_OR in 3.10
        87: "ASYNC_GEN_WRAP",  # was POP_BLOCK in 3.10
        88: "PREP_RERAISE_STAR",
        99: "SWAP",  # was ROT_N in 3.10
        113: None,  # was JUMP_ABSOLUTE in 3.10
        114: "POP_JUMP_FORWARD_IF_FALSE",  # was POP_JUMP_IF_FALSE in 3.10
        115: "POP_JUMP_FORWARD_IF_TRUE",  # was POP_JUMP_IF_TRUE in 3.10
        120: "COPY",
        121: None,  # was JUMP_IF_NOT_EXC_MATCH in 3.10
        122: "BINARY_OP",  # was SETUP_FINALLY in 3.10
        123: "SEND",
        128: "POP_JUMP_FORWARD_IF_NOT_NONE",
        129: "POP_JUMP_FORWARD_IF_NONE",  # was GEN_START in 3.10
        131: "GET_AWAITABLE",  # was CALL_FUNCTION in 3.10
        134: "JUMP_BACKWARD_NO_INTERRUPT",
        135: "MAKE_CELL",  # was LOAD_CLOSURE in 3.10
        136: "LOAD_CLOSURE",  # was LOAD_DEREF in 3.10
        137: "LOAD_DEREF",  # was STORE_DEREF in 3.10
        138: "STORE_DEREF",  # was DELETE_DEREF in 3.10
        139: "DELETE_DEREF",
        140: "JUMP_BACKWARD",
        141: None,  # was CALL_FUNCTION_KW in 3.10
        143: None,  # was SETUP_WITH in 3.10
        149: "COPY_FREE_VARS",
        151: "RESUME",
        154: None,  # was SETUP_ASYNC_WITH in 3.10
        161: None,  # was CALL_METHOD in 3.10
        166: "PRECALL",
        171: "CALL",
        172: "KW_NAMES",
        173: "POP_JUMP_BACKWARD_IF_NOT_NONE",
        174: "POP_JUMP_BACKWARD_IF_NONE",
        175: "POP_JUMP_BACKWARD_IF_FALSE",
        176: "POP_JUMP_BACKWARD_IF_TRUE",
    },
)

PYTHON_3_12_MAPPING: OpMap = _overlay_mapping(
    PYTHON_3_11_MAPPING,
    {
        3: "INTERPRETER_EXIT",
        4: "END_FOR",
        5: "END_SEND",
        10: None,  # was UNARY_POSITIVE in 3.11
        17: "RESERVED",
        26: "BINARY_SLICE",
        27: "STORE_SLICE",
        55: "CLEANUP_THROW",
        70: None,  # was PRINT_EXPR in 3.11
        82: None,  # was LIST_TO_TUPLE in 3.11
        84: None,  # was IMPORT_STAR in 3.11
        86: None,  # was YIELD_VALUE in 3.11
        87: "LOAD_LOCALS",  # was ASYNC_GEN_WRAP in 3.11
        88: None,  # was PREP_RERAISE_STAR in 3.11
        111: None,  # was JUMP_IF_FALSE_OR_POP in 3.11
        112: None,  # was JUMP_IF_TRUE_OR_POP in 3.11
        114: "POP_JUMP_IF_FALSE",  # was POP_JUMP_FORWARD_IF_FALSE in 3.11
        115: "POP_JUMP_IF_TRUE",  # was POP_JUMP_FORWARD_IF_TRUE in 3.11
        121: "RETURN_CONST",
        127: "LOAD_FAST_CHECK",
        128: "POP_JUMP_IF_NOT_NONE",  # was POP_JUMP_FORWARD_IF_NOT_NONE in 3.11
        129: "POP_JUMP_IF_NONE",  # was POP_JUMP_FORWARD_IF_NONE in 3.11
        141: "LOAD_SUPER_ATTR",
        143: "LOAD_FAST_AND_CLEAR",
        148: None,  # was LOAD_CLASSDEREF in 3.11
        150: "YIELD_VALUE",
        160: None,  # was LOAD_METHOD in 3.11
        166: None,  # was PRECALL in 3.11
        173: "CALL_INTRINSIC_1",  # was POP_JUMP_BACKWARD_IF_NOT_NONE in 3.11
        174: "CALL_INTRINSIC_2",  # was POP_JUMP_BACKWARD_IF_NONE in 3.11
        175: (
            "LOAD_FROM_DICT_OR_GLOBALS"
        ),  # was POP_JUMP_BACKWARD_IF_FALSE in 3.11
        176: "LOAD_FROM_DICT_OR_DEREF",  # was POP_JUMP_BACKWARD_IF_TRUE in 3.11
    },
)


def get_mapping(version: Tuple[int, int]) -> OpMap:
    return {
        (3, 8): PYTHON_3_8_MAPPING,
        (3, 9): PYTHON_3_9_MAPPING,
        (3, 10): PYTHON_3_10_MAPPING,
        (3, 11): PYTHON_3_11_MAPPING,
        (3, 12): PYTHON_3_12_MAPPING,
    }[version]


# ----------------------------------------------------------
# Opcode argument types

CONST = 1  # references the constant table
NAME = 2  # references the name table
JREL = 4  # relative jump
JABS = 8  # absolute jump
LOCAL = 16  # references the varnames table
FREE = 32  # references "free variable" cells
NARGS = 64  # stores number of args + kwargs


PYTHON_3_8_ARG_TYPES: OpArgs = {
    "STORE_NAME": NAME,
    "DELETE_NAME": NAME,
    "FOR_ITER": JREL,
    "STORE_ATTR": NAME,
    "DELETE_ATTR": NAME,
    "STORE_GLOBAL": NAME,
    "DELETE_GLOBAL": NAME,
    "LOAD_CONST": CONST,
    "LOAD_NAME": NAME,
    "LOAD_ATTR": NAME,
    "IMPORT_NAME": NAME,
    "IMPORT_FROM": NAME,
    "JUMP_FORWARD": JREL,
    "JUMP_IF_FALSE_OR_POP": JABS,
    "JUMP_IF_TRUE_OR_POP": JABS,
    "JUMP_ABSOLUTE": JABS,
    "POP_JUMP_IF_FALSE": JABS,
    "POP_JUMP_IF_TRUE": JABS,
    "LOAD_GLOBAL": NAME,
    "CONTINUE_LOOP": JABS,
    "SETUP_LOOP": JREL,
    "SETUP_EXCEPT": JREL,
    "SETUP_FINALLY": JREL,
    "LOAD_FAST": LOCAL,
    "STORE_FAST": LOCAL,
    "DELETE_FAST": LOCAL,
    "STORE_ANNOTATION": NAME,
    "CALL_FUNCTION": NARGS,
    "LOAD_CLOSURE": FREE,
    "LOAD_DEREF": FREE,
    "STORE_DEREF": FREE,
    "DELETE_DEREF": FREE,
    "CALL_FUNCTION_VAR": NARGS,
    "CALL_FUNCTION_KW": NARGS,
    "CALL_FUNCTION_VAR_KW": NARGS,
    "SETUP_WITH": JREL,
    "LOAD_CLASSDEREF": FREE,
    "SETUP_ASYNC_WITH": JREL,
    "LOAD_METHOD": NAME,
    "CALL_METHOD": NARGS,
    "CALL_FINALLY": JREL,
    "JUMP_IF_NOT_EXC_MATCH": JABS,
    "POP_JUMP_FORWARD_IF_FALSE": JREL,
    "POP_JUMP_FORWARD_IF_TRUE": JREL,
    "SEND": JREL,
    "POP_JUMP_FORWARD_IF_NOT_NONE": JREL,
    "POP_JUMP_FORWARD_IF_NONE": JREL,
    "JUMP_BACKWARD_NO_INTERRUPT": JREL,
    "MAKE_CELL": FREE,
    "JUMP_BACKWARD": JREL,
    "KW_NAMES": CONST,
    "POP_JUMP_BACKWARD_IF_NOT_NONE": JREL,
    "POP_JUMP_BACKWARD_IF_NONE": JREL,
    "POP_JUMP_BACKWARD_IF_FALSE": JREL,
    "POP_JUMP_BACKWARD_IF_TRUE": JREL,
}

PYTHON_3_11_ARG_TYPES: OpArgs = {
    **PYTHON_3_8_ARG_TYPES,
    "JUMP_IF_FALSE_OR_POP": JREL,
    "JUMP_IF_TRUE_OR_POP": JREL,
}

PYTHON_3_12_ARG_TYPES: OpArgs = {
    **PYTHON_3_11_ARG_TYPES,
    "POP_JUMP_IF_FALSE": JREL,
    "POP_JUMP_IF_TRUE": JREL,
    "RETURN_CONST": CONST,
    "LOAD_FAST_CHECK": LOCAL,
    "POP_JUMP_IF_NOT_NONE": JREL,
    "POP_JUMP_IF_NONE": JREL,
    "LOAD_SUPER_ATTR": NAME,
    "LOAD_FAST_AND_CLEAR": LOCAL,
    "LOAD_FROM_DICT_OR_GLOBALS": NAME,
    "LOAD_FROM_DICT_OR_DEREF": FREE,
}


def arg_type(name: str, version: Tuple[int, int]) -> Optional[int]:
    if version >= (3, 12):
        argmap = PYTHON_3_12_ARG_TYPES
    elif version >= (3, 11):
        argmap = PYTHON_3_11_ARG_TYPES
    else:
        argmap = PYTHON_3_8_ARG_TYPES
    return argmap.get(name)


# ----------------------------------------------------------
# Function names referenced by CALL_INTRINSIC_{1,2} opcodes

PYTHON_3_12_INTRINSIC_1_DESCS = [
    "INTRINSIC_1_INVALID",
    "INTRINSIC_PRINT",
    "INTRINSIC_IMPORT_STAR",
    "INTRINSIC_STOPITERATION_ERROR",
    "INTRINSIC_ASYNC_GEN_WRAP",
    "INTRINSIC_UNARY_POSITIVE",
    "INTRINSIC_LIST_TO_TUPLE",
    "INTRINSIC_TYPEVAR",
    "INTRINSIC_PARAMSPEC",
    "INTRINSIC_TYPEVARTUPLE",
    "INTRINSIC_SUBSCRIPT_GENERIC",
    "INTRINSIC_TYPEALIAS",
]

PYTHON_3_12_INTRINSIC_2_DESCS = [
    "INTRINSIC_2_INVALID",
    "INTRINSIC_PREP_RERAISE_STAR",
    "INTRINSIC_TYPEVAR_WITH_BOUND",
    "INTRINSIC_TYPEVAR_WITH_CONSTRAINTS",
    "INTRINSIC_SET_FUNCTION_TYPE_PARAMS",
]


def intrinsic_1_desc(arg: int, version: Tuple[int, int]) -> Optional[str]:
    if version >= (3, 12):
        descs = PYTHON_3_12_INTRINSIC_1_DESCS
    else:
        descs = []
    return descs[arg]


def intrinsic_2_desc(arg: int, version: Tuple[int, int]) -> Optional[str]:
    if version >= (3, 12):
        descs = PYTHON_3_12_INTRINSIC_2_DESCS
    else:
        descs = []
    return descs[arg]


# ----------------------------------------------------------
# Inline cache entry counts

PYTHON_3_11_CACHES: OpCaches = {
    "BINARY_OP": 1,
    "BINARY_SUBSCR": 4,
    "CALL": 4,
    "COMPARE_OP": 2,
    "LOAD_ATTR": 4,
    "LOAD_GLOBAL": 5,
    "LOAD_METHOD": 10,
    "PRECALL": 1,
    "STORE_ATTR": 4,
    "STORE_SUBSCR": 1,
    "UNPACK_SEQUENCE": 1,
}

PYTHON_3_12_CACHES: OpCaches = {
    **PYTHON_3_11_CACHES,
    "BINARY_SUBSCR": 1,
    "CALL": 3,
    "COMPARE_OP": 1,
    "FOR_ITER": 1,
    "LOAD_ATTR": 9,
    "LOAD_GLOBAL": 4,
    "LOAD_METHOD": 0,
    "LOAD_SUPER_ATTR": 1,
    "SEND": 1,
}


def caches(name: str, version: Tuple[int, int]) -> int:
    if version >= (3, 12):
        cachesmap = PYTHON_3_12_CACHES
    elif version >= (3, 11):
        cachesmap = PYTHON_3_11_CACHES
    else:
        cachesmap = {}
    return cachesmap.get(name, 0)



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for pycnite.pyc."""

import unittest

from . import base
from pycnite import types
from pycnite import pyc


class TestLoader(unittest.TestCase):
    """Test pyc loading."""

    def test_load(self):
        for version in base.VERSIONS:
            path = base.test_pyc("basic", version)
            with open(path, "rb") as f:
                code = pyc.load(f)
            self.assertIsInstance(code, types.CodeTypeBase)
            self.assertEqual(version, code.python_version)

    def test_loads(self):
        for version in base.VERSIONS:
            path = base.test_pyc("basic", version)
            with open(path, "rb") as f:
                data = f.read()
            code = pyc.loads(data)
            self.assertIsInstance(code, types.CodeTypeBase)
            self.assertEqual(version, code.python_version)

    def test_load_file(self):
        for version in base.VERSIONS:
            path = base.test_pyc("basic", version)
            code = pyc.load_file(path)
            self.assertIsInstance(code, types.CodeTypeBase)
            self.assertEqual(version, code.python_version)


if __name__ == "__main__":
    unittest.main()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for pycnite.bytecode."""

import unittest

from . import base
from pycnite import bytecode
from pycnite import pyc


class TestBytecode(unittest.TestCase):
    """Test bytecode reading."""

    def test_reader(self):
        path = base.test_pyc("trivial", (3, 11))
        code = pyc.load_file(path)
        opcodes = [x.name for x in bytecode.dis(code)]
        expected = [
            "RESUME",
            "LOAD_CONST",
            "STORE_NAME",
            "LOAD_CONST",
            "STORE_NAME",
            "LOAD_NAME",
            "LOAD_NAME",
            "BINARY_OP",
            "STORE_NAME",
            "LOAD_NAME",
            "LOAD_NAME",
            "BINARY_OP",
            "STORE_NAME",
            "LOAD_NAME",
            "LOAD_NAME",
            "LOAD_ATTR",
            "COMPARE_OP",
            "STORE_NAME",
            "LOAD_CONST",
            "RETURN_VALUE",
        ]
        self.assertEqual(opcodes, expected)

    def test_argval(self):
        def run(version):
            path = base.test_pyc("trivial", version)
            code = pyc.load_file(path)
            opcodes = [(x.name, x.argval) for x in bytecode.dis(code)]
            # 3.11+ bytecode starts with a RESUME opcode. Skip that.
            if version >= (3, 11):
                opcodes = opcodes[1:]
            # Grab some opcodes where we care about the argval:
            # - 0:4 are opcodes for `x = 1\ny = 2`
            # - 14:16 are the last opcodes for `b = x < y.z`
            opcodes = opcodes[0:4] + opcodes[14:16]
            expected = [
                ("LOAD_CONST", 1),
                ("STORE_NAME", "x"),
                ("LOAD_CONST", 2),
                ("STORE_NAME", "y"),
                ("LOAD_ATTR", "z"),
                ("COMPARE_OP", 0),
            ]
            self.assertEqual(opcodes, expected)

        for version in base.VERSIONS:
            run(version)

    def test_argval_for_jump_312(self):
        # In 3.12+ relative jump offsets start after CACHE entries belonging to
        # the current instruction.
        path = base.test_pyc("loop", (3, 12))
        code = pyc.load_file(path)
        opcodes = [(x.name, x.argval) for x in bytecode.dis(code)]
        opcodes = opcodes[6:7]
        expected = [
            # In 3.12 FOR_ITER is followed by a CACHE entry. If we weren't
            # accounting for that, the calculated jump target (argval) would be
            # off.
            # Verified using `dis` and godbolt.
            ("FOR_ITER", 26),
        ]
        self.assertEqual(opcodes, expected)

    def test_line_col(self):
        path = base.test_pyc("trivial", (3, 11))
        code = pyc.load_file(path)
        opcode = bytecode.dis(code)[0]
        self.assertIsNotNone(opcode.line)
        self.assertIsNotNone(opcode.endline)
        self.assertIsNotNone(opcode.col)
        self.assertIsNotNone(opcode.endcol)

    def test_extended_arg(self):
        code = bytearray([144, 10, 144, 20, 100, 1])
        ops = list(bytecode.wordcode_reader(code))
        expected_arg = 10 << 16 | 20 << 8 | 1
        expected = [bytecode.RawOpcode(0, 6, 100, expected_arg)]
        self.assertEqual(ops, expected)

    def test_generator_expression(self):
        # Check for a corner case in 3.11 generator expressions
        path = base.test_pyc("genexpr", (3, 11))
        code = pyc.load_file(path)
        d = bytecode.dis_all(code)
        genexpr = d.get_child("f").get_child("<genexpr>")
        retval = genexpr.opcodes[-1]
        self.assertEqual((retval.name, retval.line), ("RETURN_VALUE", 2))

    def test_method_calls(self):
        # Check that we assign line numbers correctly in 3.11
        # (regression test for https://github.com/google/pycnite/issues/15)
        path = base.test_pyc("method_calls", (3, 11))
        code = pyc.load_file(path)
        d = bytecode.dis_all(code)
        lines = [x.line for x in d.opcodes]
        chunks = [(0, 1), (1, 3), (2, 6), (3, 5), (4, 6), (5, 10)]
        expected = []
        for l, n in chunks:
            expected.extend([l] * n)
        self.assertEqual(lines, expected)

    def test_exception_table(self):
        def run(version):
            path = base.test_pyc("exception", version)
            code = pyc.load_file(path)
            dis = bytecode.dis_all(code)
            if version >= (3, 11):
                self.assertTrue(dis.exception_table)
            else:
                self.assertFalse(dis.exception_table)

        for version in base.VERSIONS:
            run(version)



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for pycnite.mapping."""

import unittest

from pycnite import mapping


class TestMapping(unittest.TestCase):
  """Test opcode to name mapping."""

  def run_test(self, version, op, expected):
    opmap = mapping.get_mapping(version)
    self.assertEqual(opmap.get(op), expected)

  def test_versions(self):
    """Spot checks for a few opcodes."""
    self.run_test((3, 8), 9, "NOP")
    self.run_test((3, 9), 9, "NOP")
    self.run_test((3, 9), 53, None)
    self.run_test((3, 9), 117, "IS_OP")
    self.run_test((3, 10), 99, "ROT_N")
    self.run_test((3, 10), 48, None)
    self.run_test((3, 11), 0, "CACHE")
    self.run_test((3, 11), 6, None)
    self.run_test((3, 11), 9, "NOP")
    self.run_test((3, 11), 75, "RETURN_GENERATOR")
    self.run_test((3, 12), 121, "RETURN_CONST")
    self.run_test((3, 12), 150, "YIELD_VALUE")


if __name__ == "__main__":
    unittest.main()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Load and parse .pyc files."""

import io

from typing import IO, Union

from . import magic
from . import marshal


def load(fi: IO[bytes]):
    """Parse pyc data from a stream.

    Args:
      fi: A file-like object.

    Returns:
      An instance of types.CodeTypeBase.

    Raises:
      IOError: If we can't read the file or the file is malformed.
    """
    magic_number = fi.read(2)
    python_version = magic.magic_number_to_version(magic_number)
    crlf = fi.read(2)  # cr, lf
    if crlf != b"\r\n":
        raise OSError("Malformed pyc file")
    fi.read(12)  # skip rest of header
    return marshal.loads(fi.read(), python_version)


def loads(data: Union[bytes, str]):
    """Parse pyc data from a string.

    Args:
      data: pyc data

    Returns:
      An instance of types.CodeTypeBase.
    """
    return load(io.BytesIO(data))


def load_file(path: str):
    """Parse pyc data from a file.

    Args:
      path: A file path.

    Returns:
      An instance of types.CodeTypeBase.

    Raises:
      IOError: If we can't read the file or the file is malformed.
    """
    with open(path, "rb") as f:
        return load(f)



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for pycnite.linetable."""

import itertools
import unittest

from . import base
from pycnite import linetable
from pycnite import pyc
from pycnite import types


class TestLineTable(unittest.TestCase):
    """Test linetable parsing."""

    def _get_linetable(self, testfile, version, fn=None):
        path = base.test_pyc(testfile, version)
        code = pyc.load_file(path)
        if fn is not None:
            code = code.co_consts[fn]
        lt = linetable.linetable_reader(code)
        return lt.read_all()

    def test_read(self):
        src_file = base.test_src("trivial")
        with open(src_file, "r") as f:
            src = f.readlines()
        n_lines = len(src)
        for version in base.VERSIONS:
            entries = self._get_linetable("trivial", version)
            self.assertEqual(entries[-1].line, n_lines)

    def test_flow(self):
        for version in base.VERSIONS:
            entries = self._get_linetable("flow", version)
            lines = [x.line for x in entries]
            lines = [k for k, _ in itertools.groupby(lines)]
            # Lines checked against the godbolt.org disassembler
            if version == (3, 12):
                expected = [0, 1, 2, 3, 4, 2, 1, 5, 6]
            elif version == (3, 11):
                expected = [0, 1, 2, 3, 4, 5, 6, 1, 2]
            elif version == (3, 10):
                expected = [1, 2, 3, 4, 5, 6, 1, 2]
            else:
                expected = [1, 2, 3, 4, 5, 6]
            self.assertEqual(lines, expected)

    def test_generator_311(self):
        # Check that we handle NO_COLUMN_INFO correctly in 3.11
        # (regression test for https://github.com/google/pycnite/issues/18)
        entries = self._get_linetable("generator", (3, 11), fn=0)
        self.assertEqual(len(entries), 11)
        for e in entries:
            self.assertEqual(e.line, e.endline)


class TestExceptionTable(unittest.TestCase):
    """Test exceptiontable parsing."""

    def test_basic(self):
        for version in base.VERSIONS:
            # Exception table is new in 3.11
            if version < (3, 11):
                continue
            path = base.test_pyc("exception", version)
            code = pyc.load_file(path)
            et = linetable.ExceptionTableReader(code)
            actual = et.read_all()
            entry = types.ExceptionTableEntry
            # Verified using godbolt
            if version == (3, 12):
                expected = [
                    entry(start=4, end=16, target=34, depth=0, lasti=False),
                    entry(start=22, end=30, target=82, depth=0, lasti=False),
                    entry(start=34, end=42, target=66, depth=1, lasti=True),
                    entry(start=44, end=46, target=72, depth=0, lasti=False),
                    entry(start=48, end=54, target=66, depth=1, lasti=True),
                    entry(start=56, end=58, target=72, depth=0, lasti=False),
                    entry(start=60, end=60, target=66, depth=1, lasti=True),
                    entry(start=62, end=70, target=72, depth=0, lasti=False),
                    entry(start=72, end=74, target=76, depth=1, lasti=True),
                    entry(start=82, end=84, target=90, depth=1, lasti=True),
                ]
            else:
                expected = [
                    entry(start=4, end=22, target=26, depth=0, lasti=False),
                    entry(start=24, end=24, target=66, depth=0, lasti=False),
                    entry(start=26, end=34, target=58, depth=1, lasti=True),
                    entry(start=36, end=38, target=66, depth=0, lasti=False),
                    entry(start=40, end=46, target=58, depth=1, lasti=True),
                    entry(start=48, end=50, target=66, depth=0, lasti=False),
                    entry(start=52, end=52, target=58, depth=1, lasti=True),
                    entry(start=54, end=62, target=66, depth=0, lasti=False),
                    entry(start=66, end=68, target=70, depth=1, lasti=True),
                    entry(start=78, end=86, target=92, depth=0, lasti=False),
                    entry(start=92, end=94, target=102, depth=1, lasti=True),
                ]
            self.assertEqual(actual, expected)

    def test_complex(self):
        for version in base.VERSIONS:
            # Exception table is new in 3.11
            if version < (3, 11):
                continue
            path = base.test_pyc("complex_exception", version)
            code = pyc.load_file(path)
            et = linetable.ExceptionTableReader(code.co_consts[0])
            actual = et.read_all()
            entry = types.ExceptionTableEntry
            # Verified using godbolt
            if version == (3, 12):
                expected = [
                    entry(start=8, end=18, target=138, depth=0, lasti=False),
                    entry(start=20, end=76, target=78, depth=1, lasti=True),
                    entry(start=78, end=86, target=96, depth=3, lasti=True),
                    entry(start=88, end=134, target=138, depth=0, lasti=False),
                    entry(start=138, end=154, target=242, depth=1, lasti=True),
                    entry(start=156, end=220, target=232, depth=1, lasti=True),
                    entry(start=232, end=240, target=242, depth=1, lasti=True),
                ]
            else:
                expected = [
                    entry(start=8, end=20, target=162, depth=0, lasti=False),
                    entry(start=22, end=94, target=96, depth=1, lasti=True),
                    entry(start=96, end=102, target=104, depth=3, lasti=True),
                    entry(start=104, end=108, target=162, depth=0, lasti=False),
                    entry(start=110, end=110, target=104, depth=3, lasti=True),
                    entry(start=112, end=158, target=162, depth=0, lasti=False),
                    entry(start=162, end=180, target=286, depth=1, lasti=True),
                    entry(start=182, end=264, target=276, depth=1, lasti=True),
                    entry(start=276, end=284, target=286, depth=1, lasti=True),
                ]
            self.assertEqual(actual, expected)


if __name__ == "__main__":
    unittest.main()
